
/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

package com.sunlabs.archsim.fleet.ships.stride;

import com.sunlabs.archsim.base.simulation.Command;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.OutputFunction;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.TriggerAction;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

class DoneAction extends TriggerAction {


	/* output - cycle is done, ack is an input */
	private Terminal doneTerm;
	private Terminal ackDoneTerm;
	
	/* output - internal done signal, ack is an input */
	private Terminal intDoneTerm;
	private Terminal ackIntDoneTerm;
	
	AckFunction doneFunction;
	AckAndDisable doneAndDisableFunction;
	
	public DoneAction(String name){
		super(name);
		// make a done function - send an ack out
		doneFunction= new AckFunction("Done");
		// make an AckAnd Disable function - send an ack and disable this trigger action
		doneAndDisableFunction= new AckAndDisable();
		// associate done with internal done output
		outputFunctions.put(StrideDefaults.INT_CYCLE_DONE_OUT, doneFunction);
		// associate done wioth the external done output
		outputFunctions.put(StrideDefaults.DONE, doneAndDisableFunction);
	}
	

	/** this method is invoked as a part of start trigger */
	protected void processEnableTrigger(Object params) {
		// re-enabling is not permitted, this is an error
		if (enabled) {
			director.fatalError("CountingAction "
					+ getName() + ", id= " + getID()
					+ ", was re-enabled without being disabled in-between");
		}		
	} // processEnableTrigger
	
	
	/** 
	 * Produce an void value and disable the trigger action.
	 * One of these functions is needed to disable the done action
	 * until the subsequent cycle has completed.
	 * 
	 * @author ib27688
	 *
	 */
	class AckAndDisable extends OutputFunction {
		
		public AckAndDisable() {
			super("AckAndDisable");
			/** get the function delay - for producing the next count */
			Integer fDelay= globals.intValue(StrideDefaults.STRIDE_ACK_DELAY);
			functionDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: StrideDefaults.STRIDE_ACK_DELAY_DEF;
			
		} // constructor
		
		public Data value() {
			enabled= false;
			return new Data();
		}
	} // AckAndDisable
	
	
	/**
	 * Attach a terminal.
	 * Attach only terminals that are needed for this action, ignore the rest.
	 * Self check ensures that all needed terminals are attached
	 */
	public void attach(Terminal t) {
		// get the bas ename for check
		String termName= t.getBaseName();
		// remember whether the terminal was of a bad type
		boolean badType= false;
		// remember whether the same name been added twice?
		boolean twice= false;

		if (termName.equals(StrideDefaults.ACK_DONE)) { 
			if (ackDoneTerm == null) {
				if (t instanceof InputTerminal) {
					t.attachAction(this);
					numInputs++;
					ackDoneTerm= (InputTerminal)t;
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}
		else if (termName.equals(StrideDefaults.DONE)) { 
			if (doneTerm == null) {
				if (t instanceof OutputTerminal) {
					t.attachAction(this);
					doneTerm= (OutputTerminal)t;
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}
		else if (termName.equals(StrideDefaults.INT_CYCLE_DONE_OUT)) { 
			if (intDoneTerm == null) {
				if (t instanceof OutputTerminal) {
					t.attachAction(this);
					intDoneTerm= (OutputTerminal)t;
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}
		else if (termName.equals(StrideDefaults.INT_ACK_CYCLE_DONE_OUT)) { 
			if (ackIntDoneTerm == null) {
				if (t instanceof InputTerminal) {
					t.attachAction(this);
					numInputs++;
					ackIntDoneTerm= (InputTerminal)t;
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}		
		else {
			// unknown terminal
			director.error("Terminal " 
					+ t.getName() 
					+ " cannot be attached to DoneAction "
					+ getName() + ", id= " + getID()
					+ "; terminal name not recognized");
		}
		
		// check for errors
		if (badType) {
			// the terminal was of a wrong type
			director.fatalError("DoneAction, " + getName() + ", " + getID() +  
					": Terminal " + t.getName() + " is of a bad type");
		}
		if (twice) {
			// this was the second addition of the same name
			director.fatalError("DoneAction, " + getName() + ", " + getID() 
					+ ": Terminal " + t.getName() + " cannot be attached twice");
		}
			
		
	} // attach
	
	/** accessor to startCommand */
	Command getStartCommand() {
		return getEnableTriggerCommand();
	} // getStartCommand
	
	
	/** self check - are all the terminals attached, run the self check from superclass */
	public boolean selfCheck() {
		boolean OK= true;
		boolean check;
		check= (doneTerm != null);
		OK= OK && check;
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.DONE);
		}

		check= (ackDoneTerm != null);
		OK= OK && check;		
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.ACK_DONE);
		}
		
		check= (intDoneTerm != null);
		OK= OK && check;		
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.INT_CYCLE_DONE_OUT);
		}

		check= (ackIntDoneTerm != null);
		OK= OK && check;		
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.INT_ACK_CYCLE_DONE_OUT);
		}
		
		// also run the self check on the superclass
		check= super.selfCheck();
		OK= OK && check;
		
		return OK;
	} // selfCheck

	
} // class DoneAction
